/*
 * Common clock functionality.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>

#include "cpm-common.h"
#include <math.h>



// Register access.

Cpm_regs::Cpm_regs(l4_addr_t addr, Clock_base *clocks[],
                   uint32_t exclk_freq)
: _clocks(clocks), exclk_freq(exclk_freq)
{
  _regs = new Hw::Mmio_register_block<32>(addr);
}

// Utility methods.

uint32_t
Cpm_regs::get_field(uint32_t reg, uint32_t mask, uint8_t shift)
{
  return (_regs[reg] & (mask << shift)) >> shift;
}

void
Cpm_regs::set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value)
{
  _regs[reg] = (_regs[reg] & (~(mask << shift))) | ((mask & value) << shift);
}

Clock_base *
Cpm_regs::get_clock(int num)
{
  return _clocks[num];
}



// Field methods.

uint32_t
Field::get_field(Cpm_regs &regs)
{
  if (defined)
    return regs.get_field(reg, mask, bit);
  else
    return 0;
}

void
Field::set_field(Cpm_regs &regs, uint32_t value)
{
  if (defined)
    regs.set_field(reg, mask, bit, value);
}

// Undefined field.

Field Field::undefined;



// Clock sources.

enum Clock_identifiers
Mux::get_input(int num)
{
  if (num < _num_inputs)
    return _inputs[num];
  else
    return Clock_undefined;
}

// Clock sources.

uint8_t
Source::get_source(Cpm_regs &regs)
{
  if (_source.is_defined())
    return _source.get_field(regs);
  else
    return 0;
}

void
Source::set_source(Cpm_regs &regs, uint8_t source)
{
  if (!_source.is_defined())
    return;

  _source.set_field(regs, source);
}

// Clock source frequencies.

uint32_t
Source::get_frequency(Cpm_regs &regs)
{
  // Clocks with one source yield that input frequency.

  if (get_number() == 1)
    return regs.get_clock(get_input(0))->get_frequency(regs);

  // With multiple sources, obtain the selected source for the clock.

  uint8_t source = get_source(regs);
  enum Clock_identifiers input = get_input(source);

  // Return the frequency of the source.

  if (input != Clock_undefined)
    return regs.get_clock(input)->get_frequency(regs);
  else
    return 0;
}



// Clock control.

Control_base::~Control_base()
{
}

void
Control_base::change_disable(Cpm_regs &regs)
{
  (void) regs;
}

void
Control_base::change_enable(Cpm_regs &regs)
{
  (void) regs;
}

int
Control::have_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    return !_gate.get_field(regs);
  else
    return true;
}

void
Control::start_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    _gate.set_field(regs, 0);
}

void
Control::stop_clock(Cpm_regs &regs)
{
  if (_gate.is_defined())
    _gate.set_field(regs, 1);
}

void
Control::wait_busy(Cpm_regs &regs)
{
  if (_busy.is_defined())
    while (_busy.get_field(regs));
}

void
Control::change_disable(Cpm_regs &regs)
{
  if (_change_enable.is_defined())
    _change_enable.set_field(regs, 0);
}

void
Control::change_enable(Cpm_regs &regs)
{
  if (_change_enable.is_defined())
    _change_enable.set_field(regs, 1);
}



// PLL-specific control.

int
Control_pll::have_pll(Cpm_regs &regs)
{
  return _stable.get_field(regs);
}

int
Control_pll::pll_enabled(Cpm_regs &regs)
{
  return _enable.get_field(regs);
}

int
Control_pll::pll_bypassed(Cpm_regs &regs)
{
  return _bypass.get_field(regs);
}

// Clock control.

int
Control_pll::have_clock(Cpm_regs &regs)
{
  return have_pll(regs) && pll_enabled(regs);
}

void
Control_pll::start_clock(Cpm_regs &regs)
{
  _enable.set_field(regs, 1);
  while (!have_pll(regs));
}

void
Control_pll::stop_clock(Cpm_regs &regs)
{
  _enable.set_field(regs, 0);
  while (have_pll(regs));
}

void
Control_pll::wait_busy(Cpm_regs &regs)
{
  if (pll_enabled(regs) && !pll_bypassed(regs))
    while (!have_pll(regs));
}



// Clock dividers.

Divider_base::~Divider_base()
{
}



uint32_t
Divider::get_divider(Cpm_regs &regs)
{
  if (_divider.is_defined())
    return _divider.get_field(regs) + 1;
  else
    return 1;
}

void
Divider::set_divider(Cpm_regs &regs, uint32_t divider)
{
  if (_divider.is_defined())
    _divider.set_field(regs, divider - 1);
}

// Output clock frequencies.

uint32_t
Divider::get_frequency(Cpm_regs &regs, uint32_t source_frequency)
{
  return source_frequency / get_divider(regs);
}

int
Divider::get_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  parameters[0] = get_divider(regs);
  return 1;
}

void
Divider::set_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  set_divider(regs, parameters[0]);
}



// Feedback (13-bit) multiplier.

uint32_t
Divider_pll::get_multiplier(Cpm_regs &regs)
{
  return _multiplier.get_field(regs) + 1;
}

void
Divider_pll::set_multiplier(Cpm_regs &regs, uint32_t multiplier)
{
  _multiplier.set_field(regs, multiplier - 1);
}

// Input (6-bit) divider.

uint32_t
Divider_pll::get_input_divider(Cpm_regs &regs)
{
  return _input_divider.get_field(regs) + 1;
}

void
Divider_pll::set_input_divider(Cpm_regs &regs, uint32_t divider)
{
  _input_divider.set_field(regs, divider - 1);
}

// Output (dual 3-bit) dividers.

uint32_t
Divider_pll::get_output_divider(Cpm_regs &regs)
{
  uint32_t d0 = _output_divider0.get_field(regs);
  uint32_t d1 = _output_divider1.get_field(regs);

  return d0 * d1;
}

void
Divider_pll::set_output_divider(Cpm_regs &regs, uint32_t divider)
{
  // Assert 1 as a minimum.
  // Divider 0 must be less than or equal to divider 1.

  uint32_t d0 = (uint32_t) floor(sqrt(divider ? divider : 1));
  uint32_t d1 = divider / d0;

  _output_divider0.set_field(regs, d0);
  _output_divider1.set_field(regs, d1);
}

uint32_t
Divider_pll::get_frequency(Cpm_regs &regs, uint32_t source_frequency)
{
  return (source_frequency * get_multiplier(regs)) /
         (get_input_divider(regs) * get_output_divider(regs));
}

int
Divider_pll::get_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  parameters[0] = get_multiplier(regs);
  parameters[1] = get_input_divider(regs);
  parameters[2] = get_output_divider(regs);
  return 3;
}

void
Divider_pll::set_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  set_multiplier(regs, parameters[0]);
  set_input_divider(regs, parameters[1]);
  set_output_divider(regs, parameters[2]);
}



// I2S clock divider.

uint32_t
Divider_i2s::get_multiplier(Cpm_regs &regs)
{
  return _multiplier.get_field(regs);
}

uint32_t
Divider_i2s::get_divider_N(Cpm_regs &regs)
{
  return _divider_N.get_field(regs);
}

uint32_t
Divider_i2s::get_divider_D(Cpm_regs &regs)
{
  return _divider_D.get_field(regs);
}

uint32_t
Divider_i2s::get_frequency(Cpm_regs &regs, uint32_t source_frequency)
{
  return (source_frequency * get_multiplier(regs)) /
         (get_divider_N(regs) * get_divider_D(regs));
}

int
Divider_i2s::get_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  parameters[0] = get_multiplier(regs);
  parameters[1] = get_divider_N(regs);
  parameters[2] = get_divider_D(regs);
  return 3;
}

void
Divider_i2s::set_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  // Test for N < 2M.

  if (parameters[1]  < 2 * parameters[0] )
    return;

  _multiplier.set_field(regs, parameters[0]);
  _divider_N.set_field(regs, parameters[1]);
  _divider_D.set_field(regs, parameters[2]);
}



// Clock interface.

Clock_base::~Clock_base()
{
}



// Null clock.

int
Clock_null::have_clock(Cpm_regs &regs)
{
  (void) regs;
  return false;
}

void
Clock_null::start_clock(Cpm_regs &regs)
{
  (void) regs;
}

void
Clock_null::stop_clock(Cpm_regs &regs)
{
  (void) regs;
}

// Output clock frequencies.

uint32_t
Clock_null::get_frequency(Cpm_regs &regs)
{
  (void) regs;
  return 0;
}



// Passive clock.

int
Clock_passive::have_clock(Cpm_regs &regs)
{
  (void) regs;
  return true;
}

void
Clock_passive::start_clock(Cpm_regs &regs)
{
  (void) regs;
}

void
Clock_passive::stop_clock(Cpm_regs &regs)
{
  (void) regs;
}

// Output clock frequencies.

uint32_t
Clock_passive::get_frequency(Cpm_regs &regs)
{
  // NOTE: Return the external clock frequency.

  return regs.exclk_freq;
}



// Clock control.

int
Clock_controlled::have_clock(Cpm_regs &regs)
{
  return _get_control().have_clock(regs);
}

void
Clock_controlled::start_clock(Cpm_regs &regs)
{
  _get_control().start_clock(regs);
}

void
Clock_controlled::stop_clock(Cpm_regs &regs)
{
  _get_control().stop_clock(regs);
}



// Active clock interface.

Clock_active::~Clock_active()
{
}

// Clock sources.

uint8_t
Clock_active::get_source(Cpm_regs &regs)
{
  return _source.get_source(regs);
}

void
Clock_active::set_source(Cpm_regs &regs, uint8_t source)
{
  _get_control().change_enable(regs);
  _source.set_source(regs, source);
  _get_control().wait_busy(regs);
  _get_control().change_disable(regs);
}

// Clock source frequencies.

uint32_t
Clock_active::get_source_frequency(Cpm_regs &regs)
{
  return _source.get_frequency(regs);
}

// Output clock frequencies.

uint32_t
Clock_active::get_frequency(Cpm_regs &regs)
{
  return get_source_frequency(regs);
}



// Divided clock interface.

Clock_divided_base::~Clock_divided_base()
{
}

// Output clock frequencies.

uint32_t
Clock_divided_base::get_frequency(Cpm_regs &regs)
{
  return _get_divider().get_frequency(regs, get_source_frequency(regs));
}

int
Clock_divided_base::get_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  return _get_divider().get_parameters(regs, parameters);
}

void
Clock_divided_base::set_parameters(Cpm_regs &regs, uint32_t parameters[])
{
  _get_control().change_enable(regs);
  _get_divider().set_parameters(regs, parameters);
  _get_control().wait_busy(regs);
  _get_control().change_disable(regs);
}



// PLL functionality.

Pll::~Pll()
{
}

uint32_t
Pll::get_frequency(Cpm_regs &regs)
{
  if (have_clock(regs))
  {
    if (!_control.pll_bypassed(regs))
      return _divider.get_frequency(regs, get_source_frequency(regs));
    else
      return get_source_frequency(regs);
  }
  else
    return 0;
}
